package io.tswf.groovy.bettergroovy.asm.view.generation.conflict

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.view.exception.ConflictResolveException
import io.tswf.groovy.bettergroovy.asm.view.exception.IConflictResolveExceptionBypassMarker
import io.tswf.groovy.bettergroovy.asm.view.meta.ViewConflictResolveStrategy
import io.tswf.groovy.bettergroovy.asm.view.meta.ViewMethodMetadata
import org.objectweb.asm.tree.ClassNode
import org.objectweb.asm.tree.MethodNode

@CompileStatic
class ConflictResolver {

    @Lazy
    static Map<ViewConflictResolveStrategy, IViewConflictResolveLogic> conflictResolvers = ServiceLoader.load(IViewConflictResolveLogic)
        .toMapBy { it.enumValue }
        .tap { resolvers ->
            def missingKeys = ViewConflictResolveStrategy.values()
                .stream()
                    .filter { !resolvers.containsKey(it) }
                    .toSet()

            if (missingKeys.isNotNullOrEmpty()) {
                throw new RuntimeException("No logic provided for method conflict resolve strategies: ${missingKeys}")
            }
        }


    static boolean resolveOrPassConflict(
            ViewMethodMetadata viewMethodMeta,
            ClassNode targetClass,
            MethodNode conflictingMethod
    ) {
        try {
            def resolver = conflictResolvers[viewMethodMeta.conflictStrategy]
                    .ensureNotNull { new NoSuchElementException("There is no logic for conflict resolve strategy ${viewMethodMeta.conflictStrategy?.name()}") }

            def conflictResolveContext = new ConflictResolveContext(
                    viewMethodMeta: viewMethodMeta,
                    targetClass: targetClass,
                    conflictingMethod: conflictingMethod
            )

            resolver.handleConflictResolve(conflictResolveContext)

            return conflictResolveContext.passed
        } catch (Throwable e ) {
            if (e instanceof IConflictResolveExceptionBypassMarker) {
                throw e
            }

            throw new ConflictResolveException(e, viewMethodMeta, targetClass, conflictingMethod)
        }
    }
}
