package io.tswf.groovy.bettergroovy.asm.view.reading.mappers

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.view.annotation.FieldAccessor
import io.tswf.groovy.bettergroovy.asm.view.exception.InvalidViewMethodException
import io.tswf.groovy.bettergroovy.asm.view.generation.mapper.FieldAccessType
import io.tswf.groovy.bettergroovy.asm.view.generation.mapper.FieldAccessorMapper
import io.tswf.groovy.bettergroovy.asm.view.meta.ViewMethodMetadata
import io.tswf.groovy.bettergroovy.asm.view.reading.IViewMethodMapperReader
import io.tswf.groovy.bettergroovy.asm.view.reading.ViewMethodMapperReadingContext
import io.tswf.groovy.bettergroovy.asm.view.reading.remap.ViewsRemapper
import org.objectweb.asm.Type

@CompileStatic
@AutoService(IViewMethodMapperReader)
class FieldAccessorMapperReader implements IViewMethodMapperReader {

    public static final Collection<String> PREFIXES = [ "is", "get", "set" ]
    public static final String FIELD_NAME_PARAMETER_NAME = "name"
    public static final String FIELD_DESC_PARAMETER_NAME = "desc"

    @Override
    void handleViewMethodReading(ViewMethodMapperReadingContext readingContext) {
        def viewMethod = readingContext.viewMethodMeta.viewMethod
        def viewMethodReturnType = viewMethod.returnType

        def fieldAccessorAnnotation = viewMethod.findAnnotationByType(FieldAccessor)
        def methodNameStartsFromAccessorPrefix = PREFIXES.any { prefix -> viewMethod.name.startsWith(prefix) && viewMethod.name.length() > prefix.length() }
        def methodIsFieldAccessor = fieldAccessorAnnotation != null || methodNameStartsFromAccessorPrefix

        if (methodIsFieldAccessor) {
            def parameters = viewMethod.parameterTypes

            // Проверяем параметры и возвращаемые типы методов
            if (parameters.size() > 1) {
                throw new InvalidViewMethodException("parameters count more than 1. Accessors may be getters (non-void methods with no parameters) and setters (void methods with single parameter)", readingContext.viewMethodMeta)
            }

            def isGetter = parameters.isEmpty()

            if (isGetter && viewMethodReturnType == Type.VOID_TYPE) {
                throw new InvalidViewMethodException("getter method can not return void, not a ${viewMethodReturnType.className}", readingContext.viewMethodMeta)
            }

            if (!isGetter && viewMethodReturnType != Type.VOID_TYPE) {
                throw new InvalidViewMethodException("setter method must be void type, not a ${viewMethodReturnType.className}", readingContext.viewMethodMeta)
            }

            // Вычитываем из аннотаций или из описаний методов имя и дескриптор целевого поля
            def rawFieldName = fieldAccessorAnnotation?.findValueByName(FIELD_NAME_PARAMETER_NAME, String)
                ?: isGetter
                    ? resolveFieldNameFromGetterMethod(readingContext.viewMethodMeta)
                    : resolveFieldNameFromSetterMethod(readingContext.viewMethodMeta)

            def rawFieldDesc = fieldAccessorAnnotation?.findValueByName(FIELD_DESC_PARAMETER_NAME, String)
                ?: isGetter
                    ? viewMethod.returnType.descriptor
                    : parameters.first().descriptor

            // Проверяем, что полученные дескриптор и имя валидны
            if (rawFieldName.isNullOrBlank()) {
                throw new InvalidViewMethodException("resolved target field type can not be empty or blank", readingContext.viewMethodMeta)
            }

            if (rawFieldDesc.isNullOrBlank()) {
                throw new InvalidViewMethodException("resolved target field type can not be empty or blank", readingContext.viewMethodMeta)
            }

            if (rawFieldDesc == Type.VOID_TYPE.descriptor) {
                throw new InvalidViewMethodException("resolved target field type can not be void!", readingContext.viewMethodMeta)
            }

            def (mappedFieldName, mappedFieldType) = ViewsRemapper.mapField(readingContext.viewMethodMeta.targetType, rawFieldName, Type.getType(rawFieldDesc))

            // Добавляем маппер в результат
            readingContext.result << new FieldAccessorMapper(mappedFieldName, mappedFieldType, isGetter ? FieldAccessType.GET : FieldAccessType.SET)
        }
    }

    static String resolveFieldNameFromGetterMethod(ViewMethodMetadata viewMethodMeta) {
        def methodName = viewMethodMeta.viewMethod.name
        if (methodName.startsWith("get") && methodName.length() > 3) {
            return methodName[3].toLowerCase() + methodName[4..-1]
        } else if (methodName.startsWith("is") && methodName.length() > 2) {
            return methodName[2].toLowerCase() + methodName[3..-1]
        } else {
            throw new InvalidViewMethodException("field name can't be extracted from accessor name '${methodName}' and not passed in annotation! Please, use standard 'is*' or 'get*' naming (e.g. isMyFlag or getByField) or provide additional info in annotation!", viewMethodMeta)
        }
    }

    static String resolveFieldNameFromSetterMethod(ViewMethodMetadata viewMethodMeta) {
        def methodName = viewMethodMeta.viewMethod.name
        if (methodName.startsWith("set") && methodName.length() > 3) {
            return methodName[3].toLowerCase() + methodName[4..-1]
        } else {
            throw new InvalidViewMethodException("field name can't be extracted from accessor name '${methodName}' and not passed in annotation! Please, use standard 'set*' naming (e.g. setMyFlag or setByField) or provide additional info in annotation!", viewMethodMeta)
        }
    }
}
